#!/usr/bin/env python
# encoding: utf-8
#
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2023 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2023 Florian Spreckelsen <f.spreckelsen@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.
#

import linkahead as db

from jsonschema import FormatChecker, validate, ValidationError
from pytest import raises
from unittest.mock import Mock, patch

from caosadvancedtools.json_schema_exporter import recordtype_to_json_schema as rtjs


def _mock_execute_query(query_string, unique=False, **kwargs):
    """Mock the response to queries for references."""
    all_records = db.Container()
    all_files = db.Container()
    other_type_rt = db.RecordType(name="OtherType")
    other_type_rt.add_property(name="IntegerProp", datatype=db.INTEGER, importance=db.OBLIGATORY)
    other_type_records = db.Container().extend([
        db.Record(id=100, name="otherA").add_parent(other_type_rt),
        db.Record(id=101, name="otherB").add_parent(other_type_rt),
        db.Record(id=102).add_parent(other_type_rt)
    ])
    all_records.extend(other_type_records)

    if query_string == "SELECT name, id FROM RECORD 'OtherType'":
        return other_type_records
    elif query_string == "FIND RECORDTYPE WITH name='OtherType'" and unique is True:
        return other_type_rt
    elif query_string == "SELECT name, id FROM RECORD":
        return all_records
    elif query_string == "SELECT name, id FROM FILE":
        return all_files
    else:
        return db.Container()


def test_empty_rt():

    rt = db.RecordType(name="Test", description="descr")

    schema = rtjs(rt)

    assert schema["title"] == rt.name
    assert schema["description"] == rt.description
    assert len(schema["properties"]) == 0
    assert len(schema["required"]) == 0
    assert schema["additionalProperties"] is True

    schema = rtjs(rt, additional_properties=False)

    assert schema["title"] == rt.name
    assert schema["description"] == rt.description
    assert len(schema["properties"]) == 0
    assert len(schema["required"]) == 0
    assert schema["additionalProperties"] is False

    schema = rtjs(rt, name_and_description_in_properties=True)

    assert len(schema["properties"]) == 2
    assert "name" in schema["properties"]
    assert "description" in schema["properties"]
    assert schema["properties"]["name"]["type"] == "string"
    assert schema["properties"]["description"]["type"] == "string"


def test_rt_with_scalar_props():

    rt = db.RecordType(name="Test")
    rt.add_property(name="SimpleText", datatype=db.TEXT, description="This is a simple text")
    rt.add_property(name="ObligatoryDatetime", datatype=db.DATETIME, importance=db.OBLIGATORY)
    rt.add_property(name="JustDateNoTime", datatype=db.DATETIME, description="Only dates, no times")
    rt.add_property(name="ObligatoryInteger", datatype=db.INTEGER, importance=db.OBLIGATORY)
    rt.add_property(name="Double", datatype=db.DOUBLE)
    # Suggested shouldn't influence the result in any way.
    rt.add_property(name="Boolean", datatype=db.BOOLEAN, importance=db.SUGGESTED)

    schema = rtjs(rt, additional_options_for_text_props={"JustDateNoTime": {"format": "date"}})

    assert "properties" in schema
    props = schema["properties"]
    assert len(props) == 6
    assert "required" in schema
    assert len(schema["required"]) == 2
    assert "ObligatoryDatetime" in schema["required"]
    assert "ObligatoryInteger" in schema["required"]

    assert "SimpleText" in props
    assert props["SimpleText"]["type"] == "string"
    assert "format" not in props["SimpleText"]
    assert "description" in props["SimpleText"]
    assert props["SimpleText"]["description"] == "This is a simple text"

    assert "ObligatoryDatetime" in props
    assert props["ObligatoryDatetime"]["type"] == "string"
    assert "anyOf" in props["ObligatoryDatetime"]
    assert len(props["ObligatoryDatetime"]["anyOf"]) == 2
    fmts = [fmt["format"] for fmt in props["ObligatoryDatetime"]["anyOf"]]
    assert "date" in fmts
    assert "date-time" in fmts

    assert "JustDateNoTime" in props
    assert props["JustDateNoTime"]["type"] == "string"
    assert "anyOf" not in props["JustDateNoTime"]
    assert "pattern" not in props["JustDateNoTime"]
    assert props["JustDateNoTime"]["format"] == "date"
    assert props["JustDateNoTime"]["description"] == "Only dates, no times"

    assert "ObligatoryInteger" in props
    assert props["ObligatoryInteger"]["type"] == "integer"

    assert "Double" in props
    assert props["Double"]["type"] == "number"

    assert "Boolean" in props
    assert props["Boolean"]["type"] == "boolean"

    # test validation (we turst the jsonschema.validat function, so only test
    # some more or less tricky cases with format or required).
    example = {
        "SimpleText": "something",
        "ObligatoryInteger": 23,
        "ObligatoryDatetime": "1900-01-01T12:34:56.0Z",
        "JustDateNoTime": "2023-10-13"
    }

    # We need to explicitly enable the FormatChecker, otherwise format will be
    # ignored
    # (https://python-jsonschema.readthedocs.io/en/latest/validate/#validating-formats)
    validate(example, schema, format_checker=FormatChecker())

    example = {
        "SimpleText": "something",
        "ObligatoryInteger": 23,
        "ObligatoryDatetime": "1900-01-01",
        "JustDateNoTime": "2023-10-13"
    }
    validate(example, schema, format_checker=FormatChecker())

    example = {
        "SimpleText": "something",
        "ObligatoryDatetime": "1900-01-01T12:34:56.0Z",
        "JustDateNoTime": "2023-10-13"
    }

    with raises(ValidationError):
        # required missing
        validate(example, schema, format_checker=FormatChecker())

    example = {
        "SimpleText": "something",
        "ObligatoryInteger": 23,
        "ObligatoryDatetime": "1900-01-01T12:34:56.0Z",
        "JustDateNoTime": "2023-10-13T23:59:59.123Z"
    }

    with raises(ValidationError):
        # date expected in JustDateNoTime, but datetime given
        validate(example, schema, format_checker=FormatChecker())


def test_units():

    rt = db.RecordType()
    rt.add_property(name="ScalarWithUnit", datatype=db.DOUBLE, unit="m")
    rt.add_property(name="ListWithUnit", description="This is a list.",
                    datatype=db.LIST(db.DOUBLE), unit="m")

    schema = rtjs(rt, units_in_description=True)

    props = schema["properties"]
    assert "ScalarWithUnit" in props
    assert props["ScalarWithUnit"]["type"] == "number"
    assert "description" in props["ScalarWithUnit"]
    assert props["ScalarWithUnit"]["description"] == "Unit is m."
    assert "unit" not in props["ScalarWithUnit"]

    assert "ListWithUnit" in props
    assert props["ListWithUnit"]["type"] == "array"
    assert "items" in props["ListWithUnit"]
    assert props["ListWithUnit"]["items"]["type"] == "number"
    assert "description" in props["ListWithUnit"]
    assert props["ListWithUnit"]["description"] == "This is a list. Unit is m."
    assert "unit" not in props["ListWithUnit"]

    schema = rtjs(rt, units_in_description=False)

    props = schema["properties"]
    assert "ScalarWithUnit" in props
    assert props["ScalarWithUnit"]["type"] == "number"
    assert "description" not in props["ScalarWithUnit"]
    assert "unit" in props["ScalarWithUnit"]
    assert props["ScalarWithUnit"]["unit"] == "m"

    assert "ListWithUnit" in props
    assert props["ListWithUnit"]["type"] == "array"
    assert "items" in props["ListWithUnit"]
    assert props["ListWithUnit"]["items"]["type"] == "number"
    assert "description" in props["ListWithUnit"]
    assert props["ListWithUnit"]["description"] == "This is a list."
    assert "unit" in props["ListWithUnit"]
    assert props["ListWithUnit"]["unit"] == "m"


def test_rt_with_list_props():

    rt = db.RecordType()
    rt.add_property(name="ListOfIntegers", datatype=db.LIST(
        db.INTEGER), description="List of integers")
    rt.add_property(name="ListOfPatterns", datatype=db.LIST(db.TEXT))

    schema = rtjs(rt, additional_options_for_text_props={"ListOfPatterns": {"pattern": "[A-Z]+"}})

    props = schema["properties"]

    assert "ListOfIntegers" in props
    assert props["ListOfIntegers"]["type"] == "array"
    assert "items" in props["ListOfIntegers"]
    assert props["ListOfIntegers"]["items"]["type"] == "integer"
    assert "description" not in props["ListOfIntegers"]["items"]
    assert props["ListOfIntegers"]["description"] == "List of integers"

    assert "ListOfPatterns" in props
    assert props["ListOfPatterns"]["type"] == "array"
    assert "items" in props["ListOfPatterns"]
    assert props["ListOfPatterns"]["items"]["type"] == "string"
    assert props["ListOfPatterns"]["items"]["pattern"] == "[A-Z]+"

    # Validation
    example = {
        "ListOfIntegers": [1, 2, 3],
        "ListOfPatterns": ["A", "BB", "CCC"]
    }
    validate(example, schema, format_checker=FormatChecker())

    example = {
        "ListOfIntegers": 1,
        "ListOfPatterns": ["A", "BB", "CCC"]
    }
    with raises(ValidationError):
        # No list
        validate(example, schema, format_checker=FormatChecker())

    example = {
        "ListOfIntegers": [1, 2, 3],
        "ListOfPatterns": ["A", "bb", "CCC"]
    }
    with raises(ValidationError):
        # Pattern doesn't match
        validate(example, schema, format_checker=FormatChecker())


@patch("linkahead.execute_query", new=Mock(side_effect=_mock_execute_query))
def test_rt_with_references():
    """References and lists of references to files will come later, so test if
    the errors are thrown correctly.

    """

    rt = db.RecordType()
    rt.add_property(name="RefProp", datatype=db.REFERENCE)

    schema = rtjs(rt)
    props = schema["properties"]
    assert "RefProp" in props
    assert "enum" in props["RefProp"]
    assert isinstance(props["RefProp"]["enum"], list)
    assert len(props["RefProp"]["enum"]) == len(
        db.execute_query("SELECT name, id FROM RECORD")) + len(
            db.execute_query("SELECT name, id FROM FILE"))
    assert "oneOf" not in props["RefProp"]

    rt = db.RecordType()
    rt.add_property(name="RefProp", datatype="OtherType")
    rt.add_property(name="OtherTextProp", datatype=db.TEXT)

    schema = rtjs(rt)
    props = schema["properties"]
    assert "RefProp" in props
    assert "oneOf" in props["RefProp"]
    assert len(props["RefProp"]["oneOf"]) == 2
    enum_index = 0
    if "enum" not in props["RefProp"]["oneOf"][enum_index]:
        # We can't really require the order here, so we just know that one of
        # the two elements must be the enum, the other the object.
        enum_index = 1 - enum_index
    assert "enum" in props["RefProp"]["oneOf"][enum_index]
    assert isinstance(props["RefProp"]["oneOf"][enum_index]["enum"], list)
    assert len(props["RefProp"]["oneOf"][enum_index]["enum"]) == 3
    assert "100, otherA" in props["RefProp"]["oneOf"][enum_index]["enum"]
    assert "101, otherB" in props["RefProp"]["oneOf"][enum_index]["enum"]
    assert "102" in props["RefProp"]["oneOf"][enum_index]["enum"]
    # the other element of oneOf is the OtherType object
    assert props["RefProp"]["oneOf"][1 - enum_index]["type"] == "object"
    other_props = props["RefProp"]["oneOf"][1 - enum_index]["properties"]
    assert "IntegerProp" in other_props
    assert other_props["IntegerProp"]["type"] == "integer"
    assert "required" in props["RefProp"]["oneOf"][1 - enum_index]
    assert len(props["RefProp"]["oneOf"][1 - enum_index]["required"]) == 1
    assert "IntegerProp" in props["RefProp"]["oneOf"][1 - enum_index]["required"]
    # The other prop also works as before
    assert "OtherTextProp" in props
    assert props["OtherTextProp"]["type"] == "string"

    rt = db.RecordType(name="TestType", description="Some description")
    rt.add_property(name="RefProp", datatype=db.LIST(db.REFERENCE),
                    description="I'm a list of references.")

    schema = rtjs(rt)
    assert schema["title"] == rt.name
    assert schema["description"] == rt.description
    assert "RefProp" in schema["properties"]
    ref_prop = schema["properties"]["RefProp"]
    assert ref_prop["type"] == "array"
    assert "description" in ref_prop
    assert ref_prop["description"] == "I'm a list of references."
    assert "items" in ref_prop
    items = ref_prop["items"]
    assert "enum" in items
    assert isinstance(items["enum"], list)
    assert len(items["enum"]) == len(
        db.execute_query("SELECT name, id FROM RECORD")) + len(
            db.execute_query("SELECT name, id FROM FILE"))
    assert "oneOf" not in items
    assert "description" not in items

    rt = db.RecordType()
    rt.add_property(name="RefProp", datatype=db.LIST("OtherType"))

    schema = rtjs(rt)

    rt = db.RecordType()
    rt.add_property(name="FileProp", datatype=db.FILE)

    with raises(NotImplementedError):
        schema = rtjs(rt)

    rt = db.RecordType()
    rt.add_property(name="FileProp", datatype=db.LIST(db.FILE))

    with raises(NotImplementedError):
        schema = rtjs(rt)


def test_broken():

    rt = db.RecordType()
    rt.add_property(name="something", datatype=None)

    with raises(ValueError) as ve:

        rtjs(rt)
        assert str(ve).startswith("Unknown or no property datatype.")

    rt = db.RecordType()
    rt.add_property(name="MultiProp", datatype=db.INTEGER)
    rt.add_property(name="MultiProp", datatype=db.INTEGER)

    with raises(NotImplementedError) as nie:

        rtjs(rt)
        assert "MultiProp" in str(nie)
        assert str(nie).startswith("Creating a schema for multi-properties is not specified.")
